/*
  kissback - keep it simple stupid backup
  Copyright (C) 2016 David Arendt
  Website: http://prrescue.prnet.org/index.php/kissback

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <string.h>
#include <dirent.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <fcntl.h>
#include <sys/time.h>
#include <stdlib.h>
#include <signal.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <libssh2.h>
#include <libssh2_sftp.h>
#include <lzma.h>

#ifdef __APPLE__
#include <CommonCrypto/CommonDigest.h>
#define SHA256_DIGEST_LENGTH CC_SHA256_DIGEST_LENGTH
#define SHA256_CTX CC_SHA256_CTX
#define SHA256_Init CC_SHA256_Init
#define SHA256_Update CC_SHA256_Update
#define SHA256_Final CC_SHA256_Final
#else
#include <openssl/sha.h>
#endif

#define SFTPBLOCKSIZE 131072

#define MODE_BACKUP 1
#define MODE_LISTDATES 2
#define MODE_CHECK 3
#define MODE_RESTORE 4
#define MODE_COMPACT 5

unsigned long long time2ll(time_t t)
{
  struct tm *tm;

  tm = localtime(&t);

  return ((tm->tm_year + 1900) * 10000000000ll) +
         ((tm->tm_mon  +    1) *   100000000ll) +
          (tm->tm_mday         *     1000000ll) +
          (tm->tm_hour         *       10000ll) +
          (tm->tm_min          *         100ll) +
           tm->tm_sec;
}

time_t ll2time(unsigned long long t)
{
  struct tm tm;
  unsigned long long rest;
  
  tm.tm_year = (t / 10000000000ll) - 1900;
  rest = t % 10000000000ll;
  tm.tm_mon = (rest / 100000000ll) - 1;
  rest = rest % 100000000ll;
  tm.tm_mday = (rest / 1000000ll);
  rest = rest % 1000000ll;
  tm.tm_hour = (rest / 10000ll);
  rest = rest % 10000ll;
  tm.tm_min = (rest / 100ll);
  tm.tm_sec = rest % 100ll;

  return mktime(&tm);
}

typedef struct config
{
  int mode;
  char sftpserver[1024];
  int sftpport;
  char sftpfingerprint[1024];
  char sftpuser[1024];
  char sftpkey[1024];
  char sftpdir[1024];
  char localdir[1024];
  int compression;
  char date[1024];
  int sha256check;
} CONFIG;

typedef struct readcharbuf
{
  int pos;
  int endpos;
  char buf[8192];
} READCHARBUF;

typedef struct dirinfo
{
  unsigned long ino;
  unsigned int mode;
  unsigned int uid;
  unsigned int gid;
  unsigned long long atime;
  unsigned long long mtime;
  unsigned long long ctime;
} DIRINFO;

typedef struct fileinfo
{
  char backupname[33];
  unsigned long ino;
  unsigned int mode;
  unsigned int uid;
  unsigned int gid;
  unsigned long long size;
  unsigned long long atime;
  unsigned long long mtime;
  unsigned long long ctime;
  char sha256[65];
} FILEINFO;

typedef struct fileinfolist
{
  char *name;
  char backupname[33];
  unsigned long ino;
  unsigned long long size;
  unsigned long long mtime;
  unsigned long long ctime;
  char sha256[65];
  struct fileinfolist *next;
} FILEINFOLIST;

typedef struct linkinfo
{
  char targetname[8192];
  unsigned long ino;
  unsigned int uid;
  unsigned int gid;
  unsigned long long atime;
  unsigned long long mtime;
  unsigned long long ctime;
} LINKINFO;

typedef struct fifoinfo
{
  unsigned long ino;
  unsigned int mode;
  unsigned int uid;
  unsigned int gid;
  unsigned long long atime;
  unsigned long long mtime;
  unsigned long long ctime;
} FIFOINFO;

typedef struct socketinfo
{
  unsigned long ino;
  unsigned int mode;
  unsigned int uid;
  unsigned int gid;
  unsigned long long atime;
  unsigned long long mtime;
  unsigned long long ctime;
} SOCKETINFO;

typedef struct charinfo
{
  unsigned int major;
  unsigned int minor;
  unsigned long ino;
  unsigned int mode;
  unsigned int uid;
  unsigned int gid;
  unsigned long long atime;
  unsigned long long mtime;
  unsigned long long ctime;
} CHARINFO;

typedef struct blockinfo
{
  unsigned int major;
  unsigned int minor;
  unsigned long ino;
  unsigned int mode;
  unsigned int uid;
  unsigned int gid;
  unsigned long long atime;
  unsigned long long mtime;
  unsigned long long ctime;
} BLOCKINFO;

void printdirinfo(FILE *f, DIRINFO *dirinfo)
{
  fprintf(f, "DIR %lu %o %u %u %llu %llu %llu\n", dirinfo->ino, dirinfo->mode, dirinfo->uid, dirinfo->gid, dirinfo->atime, dirinfo->mtime, dirinfo->ctime);
}

int parsedirinfo(char *info, DIRINFO *dirinfo)
{
  char type[8192];

  if (sscanf(info, "%s %lu %o %u %u %llu %llu %llu", type, &dirinfo->ino, &dirinfo->mode, &dirinfo->uid, &dirinfo->gid, &dirinfo->atime, &dirinfo->mtime, &dirinfo->ctime) == 8)
    return 1;
  else
    return 0;
}

void printfileinfo(FILE *f, FILEINFO *fileinfo)
{
  fprintf(f, "FILE %s %lu %o %u %u %llu %llu %llu %llu %s\n", fileinfo->backupname, fileinfo->ino, fileinfo->mode, fileinfo->uid, fileinfo->gid, fileinfo->size, fileinfo->atime, fileinfo->mtime, fileinfo->ctime, fileinfo->sha256);
}

int parsefileinfo(char *info, FILEINFO *fileinfo)
{
  char type[8192];

  if (sscanf(info, "%s %s %lu %o %u %u %llu %llu %llu %llu %s", type, fileinfo->backupname, &fileinfo->ino, &fileinfo->mode, &fileinfo->uid, &fileinfo->gid, &fileinfo->size, &fileinfo->atime, &fileinfo->mtime, &fileinfo->ctime, fileinfo->sha256) == 11)
    return 1;
  else
    return 0;
}

void printlinkinfo(FILE *f, LINKINFO *linkinfo)
{
  fprintf(f, "LINK %s %lu %u %u %llu %llu %llu\n", linkinfo->targetname, linkinfo->ino, linkinfo->uid, linkinfo->gid, linkinfo->atime, linkinfo->mtime, linkinfo->ctime);
}

int parselinkinfo(char *info, LINKINFO *linkinfo)
{
  char type[8192];

  if (sscanf(info, "%s %s %lu %u %u %llu %llu %llu", type, linkinfo->targetname, &linkinfo->ino, &linkinfo->uid, &linkinfo->gid, &linkinfo->atime, &linkinfo->mtime, &linkinfo->ctime) == 8)
    return 1;
  else
    return 0;
}

void printfifoinfo(FILE *f, FIFOINFO *fifoinfo)
{
  fprintf(f, "FIFO %lu %o %u %u %llu %llu %llu\n", fifoinfo->ino, fifoinfo->mode, fifoinfo->uid, fifoinfo->gid, fifoinfo->atime, fifoinfo->mtime, fifoinfo->ctime);
}

int parsefifoinfo(char *info, FIFOINFO *fifoinfo)
{
  char type[8192];

  if (sscanf(info, "%s %lu %o %u %u %llu %llu %llu", type, &fifoinfo->ino, &fifoinfo->mode, &fifoinfo->uid, &fifoinfo->gid, &fifoinfo->atime, &fifoinfo->mtime, &fifoinfo->ctime) == 8)
    return 1;
  else
    return 0;
}

void printsocketinfo(FILE *f, SOCKETINFO *socketinfo)
{
  fprintf(f, "SOCKET %lu %o %u %u %llu %llu %llu\n", socketinfo->ino, socketinfo->mode, socketinfo->uid, socketinfo->gid, socketinfo->atime, socketinfo->mtime, socketinfo->ctime);
}   

void printcharinfo(FILE *f, CHARINFO *charinfo)
{
  fprintf(f, "CHAR %u %u %lu %o %u %u %llu %llu %llu\n", charinfo->major, charinfo->minor, charinfo->ino, charinfo->mode, charinfo->uid, charinfo->gid, charinfo->atime, charinfo->mtime, charinfo->ctime);
}

int parsecharinfo(char *info, CHARINFO *charinfo)
{
  char type[8192];

  if (sscanf(info, "%s %u %u %lu %o %u %u %llu %llu %llu", type, &charinfo->major, &charinfo->minor, &charinfo->ino, &charinfo->mode, &charinfo->uid, &charinfo->gid, &charinfo->atime, &charinfo->mtime, &charinfo->ctime) == 10)
    return 1;
  else
    return 0;
}

void printblockinfo(FILE *f, BLOCKINFO *blockinfo)
{
  fprintf(f, "BLOCK %u %u %lu %o %u %u %llu %llu %llu\n", blockinfo->major, blockinfo->minor, blockinfo->ino, blockinfo->mode, blockinfo->uid, blockinfo->gid, blockinfo->atime, blockinfo->mtime, blockinfo->ctime);
}

int parseblockinfo(char *info, BLOCKINFO *blockinfo)
{
  char type[8192];

  if (sscanf(info, "%s %u %u %lu %o %u %u %llu %llu %llu", type, &blockinfo->major, &blockinfo->minor, &blockinfo->ino, &blockinfo->mode, &blockinfo->uid, &blockinfo->gid, &blockinfo->atime, &blockinfo->mtime, &blockinfo->ctime) == 10)
    return 1;
  else
    return 0;
}

char status[8192];

static void showstatus(int sig, siginfo_t *siginfo, void *context)
{
  fprintf(stderr, "%s\n", status);
}

void readline(char *line, int size, FILE *f) 
{
  int len;

  fgets(line, size, f);
  len = strlen(line);

  if (len < 1)
    return;

  if (line[len - 1] != '\n')
    return;

  line[len - 1] = 0;
}

int serverconnect(CONFIG *config, int *sockfd, LIBSSH2_SESSION **ssh, LIBSSH2_SFTP **sftp)
{
  struct hostent *host;
  struct sockaddr_in sin;
  const char *fp;
  int i;
  char fingerprint[1024];

  host = gethostbyname(config->sftpserver);

  if (host == NULL)
  {
    fprintf(stderr, "gethostbyname failed\n");

    return 0;
  }

  if (libssh2_init(0) != 0)
  {
    fprintf(stderr, "libssh2_init failed\n");

    return 0;
  }

  *sockfd = socket(AF_INET, SOCK_STREAM, 0);

  if (*sockfd < 0)
  {
    fprintf(stderr, "socket failed\n");

    return 0;
  }

  memset((struct sockaddr*)(&sin), 0, sizeof(struct sockaddr_in));
  sin.sin_family = AF_INET;
  sin.sin_port = htons(config->sftpport);
  memcpy(&sin.sin_addr.s_addr, host->h_addr_list[0], host->h_length);

  if (connect(*sockfd, (struct sockaddr*)(&sin), sizeof(struct sockaddr_in)) != 0)
  {
    fprintf(stderr, "connect failed\n");

    return 0;
  }

  *ssh = libssh2_session_init();

  if (*ssh == NULL)
  {
    fprintf(stderr, "libssh2_session_init failed\n");
    close(*sockfd);

    return 0;
  }

  libssh2_session_set_blocking(*ssh, 1);

  if (libssh2_session_handshake(*ssh, *sockfd) != 0)
  {
    fprintf(stderr, "libssh2_session_handshake failed\n");
    libssh2_session_free(*ssh);
    libssh2_exit();
    close(*sockfd);

    return 0;
  }

  fp = libssh2_hostkey_hash(*ssh, LIBSSH2_HOSTKEY_HASH_SHA1);

  if (fp == NULL)
  {
    fprintf(stderr, "libssh2_hostkey_hash failed\n");
    libssh2_session_free(*ssh);
    libssh2_exit();
    close(*sockfd);

    return 0;
  }

  fingerprint[0] = 0;

  for(i = 0; i < 20; i++)
    sprintf(fingerprint, "%s%02X", fingerprint, (unsigned char) fp[i]);

  if (config->sftpfingerprint[0] == 0)
  {
    fprintf(stderr, "WARNING: argument -sftpfingerprint missing, disabling server fingerprint check\n");
    fprintf(stderr, "server fingerprint: %s\n", fingerprint);
  }
  else if (strcmp(fingerprint, config->sftpfingerprint) != 0)
  {
    fprintf(stderr, "wrong server fingerprint: %s\n", fingerprint);
    libssh2_session_disconnect(*ssh, "disconnect");
    libssh2_session_free(*ssh);
    libssh2_exit();
    close(*sockfd);

    return 0;
  }

  if (libssh2_userauth_publickey_fromfile(*ssh, config->sftpuser, NULL, config->sftpkey, NULL) != 0)
  {
    fprintf(stderr, "libssh2_userauth_publickey_fromfile failed\n");
    libssh2_session_disconnect(*ssh, "disconnect");
    libssh2_session_free(*ssh);
    libssh2_exit();
    close(*sockfd);

    return 0;
  }

  *sftp = libssh2_sftp_init(*ssh);

  if (*sftp == NULL)
  {
    fprintf(stderr, "libssh2_sftp_init failed\n");
    libssh2_session_disconnect(*ssh, "disconnect");
    libssh2_session_free(*ssh);
    libssh2_exit();
    close(*sockfd);

    return 0;
  }

  return 1;
}

void serverdisconnect(int sockfd, LIBSSH2_SESSION *ssh, LIBSSH2_SFTP *sftp)
{
  libssh2_sftp_shutdown(sftp);
  libssh2_session_disconnect(ssh, "disconnect");
  libssh2_session_free(ssh);
  libssh2_exit();
  close(sockfd);
}

int sftpreadn(LIBSSH2_SFTP_HANDLE *file, char *buf, int count)
{
  int len;
  int pos = 0;

  for (;;)
  {
    len = libssh2_sftp_read(file, &buf[pos], count - pos);
    
    if (len <= 0)
      return pos;

    pos += len;

    if ((count - pos) == 0)
      return pos;
  }
}

int sftpreadchar(LIBSSH2_SFTP_HANDLE *file, READCHARBUF *buf, char *c)
{
 if (buf->pos == buf->endpos)
 {
   buf->pos = 0;
   buf->endpos = sftpreadn(file, buf->buf, sizeof(buf->buf));

   if (buf->endpos <= 0)
     return 0;
 }

 *c = buf->buf[buf->pos];
 buf->pos++;

 return 1;
}

int sftpreadline(LIBSSH2_SFTP_HANDLE *file, READCHARBUF *buf, char *value, int count)
{
  int i;
  char c;

  for (i = 0; i < count; i++)
  {
    if (!sftpreadchar(file, buf, &c))
    {
      if (i == 0)
        return 0;

      value[i] = 0;

      return 1;
    }

    if (c == '\n')
    {
      value[i] = 0;

      return 1;
    }

    value[i] = c;
  }

  return 0;
}

FILEINFOLIST *loadfileinfolist(LIBSSH2_SFTP_HANDLE *file)
{
  READCHARBUF buf;
  char name[8192];
  char info[8192];
  FILEINFO fileinfo;
  FILEINFOLIST *listelement;
  FILEINFOLIST *list = NULL;
  FILEINFOLIST *previouslistelement;

  memset(&buf, 0, sizeof(READCHARBUF));

  while (sftpreadline(file, &buf, name, 8192))
  {
    if (!sftpreadline(file, &buf, info, 8182))
      return NULL;

    if (strncmp(info, "FILE", 4) != 0)
      continue;

    if (!parsefileinfo(info, &fileinfo))
      return NULL;

    if ((listelement = malloc(sizeof(FILEINFOLIST))) == NULL)
      return NULL;

    if ((listelement->name = malloc(strlen(name) + 1)) == NULL)
      return NULL;

    strcpy(listelement->name, name);
    strcpy(listelement->backupname, fileinfo.backupname);
    listelement->ino = fileinfo.ino;
    listelement->size = fileinfo.size;
    listelement->mtime = fileinfo.mtime;
    listelement->ctime = fileinfo.ctime;
    strcpy(listelement->sha256, fileinfo.sha256);
    listelement->next = NULL;

    if (list == NULL)
      list = listelement;
    else 
      previouslistelement->next = listelement;

    previouslistelement = listelement;
  }

  return list;
}

void freefileinfolist(FILEINFOLIST *list)
{
  FILEINFOLIST *listelement;
  FILEINFOLIST *nextlistelement;

  if (list == NULL)
    return;

  listelement = list;

  do
  {
    nextlistelement = listelement->next;
    free(listelement->name);
    free(listelement);
    listelement = nextlistelement;
  }
  while (listelement != NULL);
}

int copytosftp(int source, LIBSSH2_SFTP_HANDLE *file, unsigned long long size)
{
  unsigned long long rest = size;
  char buf[SFTPBLOCKSIZE];
  int len;
  char *bufptr;
  int writelen;
  int writtenlen;

  while (rest > 0)
  { 
    if (rest < SFTPBLOCKSIZE) 
      len = read(source, buf, rest);
    else
      len = read(source, buf, SFTPBLOCKSIZE);
    
    if ((len == -1) && (errno == EINTR))
    { 
      fprintf(stderr, "resume\n");
      if (rest < SFTPBLOCKSIZE) 
        len = read(source, buf, rest);
      else
        len = read(source, buf, SFTPBLOCKSIZE);
    }
    
    if (len < 0)
      return 0;
    
    bufptr = buf;
    writelen = len;
    
    while (writelen > 0)
    { 
      writtenlen = libssh2_sftp_write(file, bufptr, writelen);
      
      if (writtenlen < 0)
        return 0;
      
      bufptr += writtenlen;
      writelen -= writtenlen;
    }
    
    rest -= len;
  }

  return 1;
}

int copyfromsftp(LIBSSH2_SFTP_HANDLE *file, int destination, unsigned long long size)
{
  unsigned long long rest = size;
  char buf[SFTPBLOCKSIZE];
  int len;
  int i;
  int i1;

  while (rest > 0)
  {
    if (rest < SFTPBLOCKSIZE)
      len = libssh2_sftp_read(file, buf, rest);
    else
      len = libssh2_sftp_read(file, buf, SFTPBLOCKSIZE);

    if (len <= 0)
      return 0;

    i = write(destination, buf, len);

    if (i < len)
    {
      if ((i == -1) && (errno == EINTR))
      {
        fprintf(stderr, "resume\n");
        i = write(destination, buf, len);
      }
      else
      {
        fprintf(stderr, "short write resume\n");
        i1 = write(destination, buf + i, len - i);
        if (i1 == -1)
          i = i1;
        else
          i += i1;
      }
    }

    if (i != len)
      return 0;

    rest -= len;
  }

  return 1;
}

int copytosftpsha256(int source, LIBSSH2_SFTP_HANDLE *file, unsigned long long size, char *sha256)
{
  unsigned char hash[SHA256_DIGEST_LENGTH];
  SHA256_CTX sha256ctx;
  unsigned long long rest = size;
  char buf[SFTPBLOCKSIZE];
  int len;
  char *bufptr;
  int writelen;
  int writtenlen;
  int i;

  SHA256_Init(&sha256ctx);

  while (rest > 0)
  {
    if (rest < SFTPBLOCKSIZE)
      len = read(source, buf, rest);
    else
      len = read(source, buf, SFTPBLOCKSIZE);

    if ((len == -1) && (errno == EINTR))
    {
      fprintf(stderr, "resume\n");
      if (rest < SFTPBLOCKSIZE)
        len = read(source, buf, rest);
      else
        len = read(source, buf, SFTPBLOCKSIZE);
    }

    if (len < 0)
      return 0;

    bufptr = buf;
    writelen = len;

    while (writelen > 0)
    {
      writtenlen = libssh2_sftp_write(file, bufptr, writelen);

      if (writtenlen == LIBSSH2_ERROR_TIMEOUT)
      {
        fprintf(stderr, "resume\n");
        writtenlen = libssh2_sftp_write(file, bufptr, writelen);
      }

      if (writtenlen < 0)
        return 0;

      bufptr += writtenlen;
      writelen -= writtenlen;
    }
    
    SHA256_Update(&sha256ctx, buf, len);
    rest -= len;
  }

  SHA256_Final(hash, &sha256ctx);

  for (i = 0; i < SHA256_DIGEST_LENGTH; i++)
    sprintf(sha256 + (i * 2), "%02x", hash[i]);

  sha256[64] = 0;

  return 1;
}

int copylzmatosftpsha256(CONFIG *config, int source, LIBSSH2_SFTP_HANDLE *file, unsigned long long size, char *sha256)
{
  unsigned char hash[SHA256_DIGEST_LENGTH];
  SHA256_CTX sha256ctx;
  lzma_mt mt = { .flags = 0, .block_size = 0, .timeout = 0, .preset = config-> compression, .filters = NULL, .check = LZMA_CHECK_CRC64 };
  lzma_stream strm = LZMA_STREAM_INIT;
  unsigned long long rest = size;
  char buf[SFTPBLOCKSIZE];
  char outbuf[SFTPBLOCKSIZE];
  int len;
  char *bufptr;
  lzma_action action = LZMA_RUN;
  lzma_ret ret;
  int writelen;
  int writtenlen;
  int i;
  unsigned long long uncompressedblocks = 0;
  unsigned long long compressedblocks = 0;

  SHA256_Init(&sha256ctx);

  mt.threads = lzma_cputhreads();

  if (mt.threads == 0)
    mt.threads = 1;

  if (lzma_stream_encoder_mt(&strm, &mt) != LZMA_OK)
    return 0;

  strm.next_in = NULL;
  strm.avail_in = 0;
  strm.next_out = (uint8_t *) outbuf;
  strm.avail_out = SFTPBLOCKSIZE;

  for (;;)
  {
    if ((strm.avail_in == 0) && (rest > 0))
    {
      if (rest < SFTPBLOCKSIZE)
        len = read(source, buf, rest);
      else
        len = read(source, buf, SFTPBLOCKSIZE);

      if ((len == -1) && (errno == EINTR))
      {
        fprintf(stderr, "resume\n");
        if (rest < SFTPBLOCKSIZE)
          len = read(source, buf, rest);
        else
          len = read(source, buf, SFTPBLOCKSIZE);
      }

      if (len < 0)
      {
        lzma_end(&strm);

        return 0;
      }

      SHA256_Update(&sha256ctx, buf, len);
      rest -= len;

      strm.next_in = (uint8_t *) buf;
      strm.avail_in = len;

      if (rest == 0)
        action = LZMA_FINISH;

      uncompressedblocks++;
    }

    ret = lzma_code(&strm, action);

    if ((ret != LZMA_OK) && (ret != LZMA_STREAM_END))
    {
      lzma_end(&strm);

      return 0;
    }

    if ((strm.avail_out == 0) || (ret == LZMA_STREAM_END))
    {
      bufptr = outbuf;
      writelen = SFTPBLOCKSIZE - strm.avail_out;

      while (writelen > 0)
      {
        writtenlen = libssh2_sftp_write(file, bufptr, writelen);

        if (writtenlen == LIBSSH2_ERROR_TIMEOUT)
        {
          fprintf(stderr, "resume\n");
          writtenlen = libssh2_sftp_write(file, bufptr, writelen);
        }

        if (writtenlen < 0)
        {
          lzma_end(&strm);

          return 0;
        }

        bufptr += writtenlen;
        writelen -= writtenlen;
      }

      if (ret == LZMA_STREAM_END)
        break;

      strm.next_out = (uint8_t *) outbuf;
      strm.avail_out = SFTPBLOCKSIZE;

      compressedblocks++;

      if ((uncompressedblocks == 5000) && (compressedblocks > (4000 - (mt.threads * 192))))
      {
        lzma_end(&strm);

        return -1;
      }
    }
  }

  lzma_end(&strm);
  SHA256_Final(hash, &sha256ctx);

  for (i = 0; i < SHA256_DIGEST_LENGTH; i++)
    sprintf(sha256 + (i * 2), "%02x", hash[i]);

  sha256[64] = 0;

  return 1;
}

int copyfromsftpsha256(LIBSSH2_SFTP_HANDLE *file, int destination, unsigned long long size, char *sha256)
{
  unsigned char hash[SHA256_DIGEST_LENGTH];
  SHA256_CTX sha256ctx;
  unsigned long long rest = size;
  char buf[SFTPBLOCKSIZE];
  int len;
  int i;
  int i1;

  SHA256_Init(&sha256ctx);

  while (rest > 0)
  {
    if (rest < SFTPBLOCKSIZE)
      len = libssh2_sftp_read(file, buf, rest);
    else
      len = libssh2_sftp_read(file, buf, SFTPBLOCKSIZE);

    if (len == LIBSSH2_ERROR_TIMEOUT)
    {
      fprintf(stderr, "resume\n");

      if (rest < SFTPBLOCKSIZE)
        len = libssh2_sftp_read(file, buf, rest);
      else
        len = libssh2_sftp_read(file, buf, SFTPBLOCKSIZE);
    }

    if (len <= 0)
      return 0;

    i = write(destination, buf, len);

    if (i < len)
    {
      if ((i == -1) && (errno == EINTR))
      {
        fprintf(stderr, "resume\n");
        i = write(destination, buf, len);
      }
      else
      {
        fprintf(stderr, "short write resume\n");
        i1 = write(destination, buf + i, len - i);
        if (i1 == -1)
          i = i1;
        else
          i += i1;
      }
    }

    if (i != len)
      return 0;

    SHA256_Update(&sha256ctx, buf, len);
    rest -= len;
  }

  SHA256_Final(hash, &sha256ctx);

  for (i = 0; i < SHA256_DIGEST_LENGTH; i++)
    sprintf(sha256 + (i * 2), "%02x", hash[i]);

  sha256[64] = 0;

  return 1;
}

int copylzmafromsftpsha256(LIBSSH2_SFTP_HANDLE *file, int destination, char *sha256)
{
  unsigned char hash[SHA256_DIGEST_LENGTH];
  SHA256_CTX sha256ctx;
  lzma_stream strm = LZMA_STREAM_INIT;
  char buf[SFTPBLOCKSIZE];
  char outbuf[SFTPBLOCKSIZE];
  int len = 1;
  lzma_action action = LZMA_RUN;
  lzma_ret ret;
  int wlen;
  int i;
  int i1;

  SHA256_Init(&sha256ctx);

  if (lzma_stream_decoder(&strm, UINT64_MAX, LZMA_CONCATENATED) != LZMA_OK)
    return 0;

  strm.next_in = NULL;
  strm.avail_in = 0; 
  strm.next_out = (uint8_t *) outbuf;
  strm.avail_out = SFTPBLOCKSIZE;

  for (;;)
  {
    if ((strm.avail_in == 0) && (len != 0))
    {
      len = libssh2_sftp_read(file, buf, SFTPBLOCKSIZE);

      if (len == LIBSSH2_ERROR_TIMEOUT)
      {
        fprintf(stderr, "resume\n");

        len = libssh2_sftp_read(file, buf, SFTPBLOCKSIZE);
      }

      if (len < 0)
      {
        lzma_end(&strm);

        return 0;
      }

      strm.next_in = (uint8_t *) buf;
      strm.avail_in = len;

      if (len == 0)
        action = LZMA_FINISH;
    }

    ret = lzma_code(&strm, action);

    if ((ret != LZMA_OK) && (ret != LZMA_STREAM_END))
    {
      lzma_end(&strm);

      return 0;
    }

    if ((strm.avail_out == 0) || (ret == LZMA_STREAM_END))
    {
      wlen = SFTPBLOCKSIZE - strm.avail_out;
      i = write(destination, outbuf, wlen);

      if (i < wlen)
      {
        if ((i == -1) && (errno == EINTR))
        {
          fprintf(stderr, "resume\n");
          i = write(destination, outbuf, wlen);
        }
        else
        {
          fprintf(stderr, "short write resume\n");
          i1 = write(destination, outbuf + i, wlen - i);
          if (i1 == -1)
            i = i1;
          else
            i += i1;
        }
      }

      if (i != wlen)
      {
        lzma_end(&strm);

        return 0;
      }

      SHA256_Update(&sha256ctx, outbuf, wlen);

      if (ret == LZMA_STREAM_END)
        break;

      strm.next_out = (uint8_t *) outbuf;
      strm.avail_out = SFTPBLOCKSIZE;
    }
  }

  lzma_end(&strm);
  SHA256_Final(hash, &sha256ctx);

  for (i = 0; i < SHA256_DIGEST_LENGTH; i++)
    sprintf(sha256 + (i * 2), "%02x", hash[i]);

  sha256[64] = 0;

  return 1;
}

int calcsha256fromsftp(LIBSSH2_SFTP_HANDLE *file, long long size, char *sha256)
{
  unsigned char hash[SHA256_DIGEST_LENGTH];
  SHA256_CTX sha256ctx;
  long long rest = size;
  char buf[SFTPBLOCKSIZE];
  int len;
  int i;

  SHA256_Init(&sha256ctx);

  while (rest > 0)
  {
    if (rest < SFTPBLOCKSIZE)
      len = libssh2_sftp_read(file, buf, rest);
    else
      len = libssh2_sftp_read(file, buf, SFTPBLOCKSIZE);

    if (len == LIBSSH2_ERROR_TIMEOUT)
    {
      fprintf(stderr, "resume\n");

      if (rest < SFTPBLOCKSIZE)
        len = libssh2_sftp_read(file, buf, rest);
      else
        len = libssh2_sftp_read(file, buf, SFTPBLOCKSIZE);
    }

    if (len <= 0)
      return 0;

    SHA256_Update(&sha256ctx, buf, len);
    rest -= len;
  }

  SHA256_Final(hash, &sha256ctx);

  for (i = 0; i < SHA256_DIGEST_LENGTH; i++)
    sprintf(sha256 + (i * 2), "%02x", hash[i]);

  sha256[64] = 0;

  return 1;
}

int calcsha256lzmafromsftp(LIBSSH2_SFTP_HANDLE *file, char *sha256)
{
  unsigned char hash[SHA256_DIGEST_LENGTH];
  SHA256_CTX sha256ctx;
  lzma_stream strm = LZMA_STREAM_INIT;
  char buf[SFTPBLOCKSIZE];
  char outbuf[SFTPBLOCKSIZE];
  int len = 1;
  lzma_action action = LZMA_RUN;
  lzma_ret ret;
  int i;

  SHA256_Init(&sha256ctx);

  if (lzma_stream_decoder(&strm, UINT64_MAX, LZMA_CONCATENATED) != LZMA_OK)
    return 0;

  strm.next_in = NULL;
  strm.avail_in = 0; 
  strm.next_out = (uint8_t *) outbuf;
  strm.avail_out = SFTPBLOCKSIZE;

  for (;;)
  {
    if ((strm.avail_in == 0) && (len != 0))
    {
      len = libssh2_sftp_read(file, buf, SFTPBLOCKSIZE);

      if (len == LIBSSH2_ERROR_TIMEOUT)
      {
        fprintf(stderr, "resume\n");

        len = libssh2_sftp_read(file, buf, SFTPBLOCKSIZE);
      }

      if (len < 0)
      {
        lzma_end(&strm);

        return 0;
      }

      strm.next_in = (uint8_t *) buf;
      strm.avail_in = len;

      if (len == 0)
        action = LZMA_FINISH;
    }

    ret = lzma_code(&strm, action);

    if ((ret != LZMA_OK) && (ret != LZMA_STREAM_END))
    {
      lzma_end(&strm);

      return 0;
    }

    if ((strm.avail_out == 0) || (ret == LZMA_STREAM_END))
    {
      SHA256_Update(&sha256ctx, outbuf, SFTPBLOCKSIZE - strm.avail_out);

      if (ret == LZMA_STREAM_END)
        break;

      strm.next_out = (uint8_t *) outbuf;
      strm.avail_out = SFTPBLOCKSIZE;
    }
  }

  lzma_end(&strm);
  SHA256_Final(hash, &sha256ctx);

  for (i = 0; i < SHA256_DIGEST_LENGTH; i++)
    sprintf(sha256 + (i * 2), "%02x", hash[i]);

  sha256[64] = 0;

  return 1;
}

int endswith(char *s, char *suffix)
{
  int len = strlen(s);
  int suffixlen = strlen(suffix);

  if (suffixlen > len)
    return 0;

  return (strcmp(s + len - suffixlen, suffix) == 0);
}

int findlastlist(LIBSSH2_SFTP *sftp, CONFIG *config, char *list)
{
  LIBSSH2_SFTP_HANDLE *dir;
  char name[1024];
  LIBSSH2_SFTP_ATTRIBUTES attributes;
  unsigned long long t;
  unsigned long long maxt = 0;

  dir = libssh2_sftp_opendir(sftp, config->sftpdir);

  if (dir == NULL)
    return 0;

  while ((libssh2_sftp_readdir(dir, name, 1000, &attributes)) > 0)
  {
    if (endswith(name , ".lst"))
    {
      name[14] = 0;
      sscanf(name, "%llu", &t);

      if (t > maxt)
        maxt = t;
    }
  }

  if (libssh2_sftp_closedir(dir) != 0)
    return 0;

  if (maxt == 0)
  {
    list[0] = 0;

    return 1;
  }

  sprintf(list, "%s/%llu.lst", config->sftpdir, maxt);

  return 1;
}

int backupdir(CONFIG *config, char *sourcedir, char *source, dev_t device, FILEINFOLIST *sourcelist, unsigned long long t, LIBSSH2_SFTP *sftp, char *sftpdir, FILE *list, unsigned long *dirnum, unsigned long *filenum)
{
  DIR *dir;
  struct dirent *entry;
  char name[8192];
  struct stat st;
  int ret;
  char destname[8192];
  DIRINFO dirinfo;
  FILEINFO fileinfo;
  FILEINFOLIST *sourcelistelement;
  FILEINFOLIST *previoussourcelistelement;
  LINKINFO linkinfo;
  FIFOINFO fifoinfo;
  SOCKETINFO socketinfo;
  CHARINFO charinfo;
  BLOCKINFO blockinfo;
  int sourcefd;
  LIBSSH2_SFTP_HANDLE *file;

  if ((dir = opendir(source)) == NULL)
  {
    perror(source);

    return 1;
  }

  while ((entry=readdir(dir)) != NULL)
  {
    if (strcmp(entry->d_name, ".") == 0)
      continue;

    if (strcmp(entry->d_name, "..") == 0)
      continue;

    if ((strlen(source) + strlen(entry->d_name) + 10) > 8192)
    {
      fprintf(stderr, "name too long\n");

      return 1;
    }

    strcpy(name, source);
    strcat(name, "/");
    strcat(name, entry->d_name);
    strcpy(status, name);

    if (lstat(name, &st) < 0)
    {
      perror(name);

      return 1;
    }

    if (st.st_dev != device)
      continue;

    fprintf(list, "%s\n", name + strlen(sourcedir) + 1);

    switch (st.st_mode & S_IFMT)
    {
      case S_IFDIR  : dirinfo.ino = st.st_ino;
                      dirinfo.mode = st.st_mode & 07777;
                      dirinfo.uid = st.st_uid;
                      dirinfo.gid = st.st_gid;
                      dirinfo.atime = time2ll(st.st_atime);
                      dirinfo.mtime = time2ll(st.st_mtime);
                      dirinfo.ctime = time2ll(st.st_ctime);
                      printdirinfo(list, &dirinfo);

                      if ((ret = backupdir(config, sourcedir, name, device, sourcelist, t, sftp, sftpdir, list, dirnum, filenum)) > 0)
                        return ret;

                      break;

      case S_IFREG  : fileinfo.ino = st.st_ino;
                      fileinfo.mode = st.st_mode & 07777;
                      fileinfo.uid = st.st_uid;
                      fileinfo.gid = st.st_gid;
                      fileinfo.size = st.st_size;
                      fileinfo.atime = time2ll(st.st_atime);
                      fileinfo.mtime = time2ll(st.st_mtime);
                      fileinfo.ctime = time2ll(st.st_ctime);

                      if (sourcelist != NULL)
                      {
                        previoussourcelistelement = NULL;

                        for (sourcelistelement = sourcelist; sourcelistelement->next != NULL; sourcelistelement = sourcelistelement->next)
                        {
                          if (strcmp(sourcelistelement->name, name + strlen(sourcedir) + 1) == 0)
                            break;

                          previoussourcelistelement = sourcelistelement;
			}

                        if (strcmp(sourcelistelement->name, name + strlen(sourcedir) + 1) == 0)
                        {
                          if ((sourcelistelement->ino == fileinfo.ino) &&
                              (sourcelistelement->size == fileinfo.size) &&
                              (sourcelistelement->mtime == fileinfo.mtime) &&
                              (sourcelistelement->ctime == fileinfo.ctime))
                          {
                            strcpy(fileinfo.backupname, sourcelistelement->backupname);
			    strcpy(fileinfo.sha256, sourcelistelement->sha256);
                            printfileinfo(list, &fileinfo);

                            if (previoussourcelistelement != NULL)
                            {
                              previoussourcelistelement->next = sourcelistelement->next;
                              free(sourcelistelement->name);
                              free(sourcelistelement);
                            }

                            break;
                          }

                          if (previoussourcelistelement != NULL)
                          {
                            previoussourcelistelement->next = sourcelistelement->next;
                            free(sourcelistelement->name);
                            free(sourcelistelement);
                          }
                        }
                      }

                      if (*filenum == 1000)
                      {
                        (*dirnum)++;
                        *filenum = 0;
                      }

                      if (*filenum == 0)
                      {
                        sprintf(destname, "%s/%06lu", sftpdir, *dirnum);

                        if (libssh2_sftp_mkdir(sftp, destname, LIBSSH2_SFTP_S_IRUSR|LIBSSH2_SFTP_S_IWUSR|LIBSSH2_SFTP_S_IXUSR) != 0)
                        {
                          fprintf(stderr, "libssh2_sftp_mkdir failed\n");

                          return 1;
                        }
                      }

                      if ((config->compression > 0) && (fileinfo.size > 0))
                        sprintf(destname, "%s/%06lu/%03lu.dat.xz", sftpdir, *dirnum, *filenum);
                      else
                        sprintf(destname, "%s/%06lu/%03lu.dat", sftpdir, *dirnum, *filenum);

                      if ((sourcefd = open(name, O_RDONLY)) < 0)
                      {
                        perror(name);

                        return 1;
                      }

                      file = libssh2_sftp_open(sftp, destname, LIBSSH2_FXF_WRITE|LIBSSH2_FXF_CREAT, LIBSSH2_SFTP_S_IRUSR|LIBSSH2_SFTP_S_IWUSR);

                      if (file == NULL)
                      {
                        fprintf(stderr, "libssh2_sftp_open failed\n");

                        return 1;
                      }

                      if ((config->compression > 0) && (fileinfo.size > 0))
                      {
                        ret = copylzmatosftpsha256(config, sourcefd, file, st.st_size, fileinfo.sha256);

                        if (ret == 0)
                        {
                          fprintf(stderr, "unable to copy %s to %s\n", name, destname);
                          libssh2_sftp_close(file);

                          return 1;
                        }
                        else if (ret == -1)
                        {
                          fprintf(stderr, "bad compression ratio for %s, switching to uncompressed\n", name);

                          if (libssh2_sftp_close(file) != 0)
                          {
                            fprintf(stderr, "libssh2_sftp_close failed\n");

                            return 1;
                          }

                          close(sourcefd);

                          if (libssh2_sftp_unlink(sftp, destname) != 0)
                          {
                            fprintf(stderr, "libssh2_sftp_unlink failed\n");

                            return 1;
                          }

                          sprintf(destname, "%s/%06lu/%03lu.dat", sftpdir, *dirnum, *filenum);

                          if ((sourcefd = open(name, O_RDONLY)) < 0)
                          {
                            perror(name);

                            return 1;
                          }

                          file = libssh2_sftp_open(sftp, destname, LIBSSH2_FXF_WRITE|LIBSSH2_FXF_CREAT, LIBSSH2_SFTP_S_IRUSR|LIBSSH2_SFTP_S_IWUSR);

                          if (file == NULL)
                          {
                            fprintf(stderr, "libssh2_sftp_open failed\n");

                            return 1;
                          }

                          if (!copytosftpsha256(sourcefd, file, st.st_size, fileinfo.sha256))
                          {
                            fprintf(stderr, "unable to copy %s to %s\n", name, destname);
                            libssh2_sftp_close(file);

                            return 1;
                          }
                        }
                      }
                      else
                      {
                        if (!copytosftpsha256(sourcefd, file, st.st_size, fileinfo.sha256))
                        {
                          fprintf(stderr, "unable to copy %s to %s\n", name, destname);
                          libssh2_sftp_close(file);

                          return 1;
                        }
                      }

                      if (libssh2_sftp_close(file) != 0)
                      {
                        fprintf(stderr, "libssh2_sftp_close failed\n");

                        return 1;
                      }

                      close(sourcefd);

                      if ((config->compression > 0) && (fileinfo.size > 0) && (ret == 1))
                        sprintf(fileinfo.backupname, "%llu/%06lu/%03lu.dat.xz", t, *dirnum, *filenum);
                      else
                        sprintf(fileinfo.backupname, "%llu/%06lu/%03lu.dat", t, *dirnum, *filenum);

                      printfileinfo(list, &fileinfo);

                      (*filenum)++;

                      break;

      case S_IFLNK  : if ((ret = readlink(name, linkinfo.targetname, 8182)) < 0)
		      {
                        perror(name);

                        return 1;
                      }

                      linkinfo.targetname[ret] = 0;

                      linkinfo.ino = st.st_ino;
                      linkinfo.uid = st.st_uid;
                      linkinfo.gid = st.st_gid;
                      linkinfo.atime = time2ll(st.st_atime);
                      linkinfo.mtime = time2ll(st.st_mtime);
                      linkinfo.ctime = time2ll(st.st_ctime);
                      printlinkinfo(list, &linkinfo);

                      break;

      case S_IFIFO  : fifoinfo.ino = st.st_ino;
                      fifoinfo.mode = st.st_mode & 07777;
                      fifoinfo.uid = st.st_uid;
                      fifoinfo.gid = st.st_gid;
                      fifoinfo.atime = time2ll(st.st_atime);
                      fifoinfo.mtime = time2ll(st.st_mtime);
                      fifoinfo.ctime = time2ll(st.st_ctime);
                      printfifoinfo(list, &fifoinfo);

                      break;

      case S_IFSOCK : socketinfo.ino = st.st_ino;
                      socketinfo.mode = st.st_mode & 07777;
                      socketinfo.uid = st.st_uid;
                      socketinfo.gid = st.st_gid;
                      socketinfo.atime = time2ll(st.st_atime);
                      socketinfo.mtime = time2ll(st.st_mtime);
                      socketinfo.ctime = time2ll(st.st_ctime);
                      printsocketinfo(list, &socketinfo);

                      break;

      case S_IFCHR  : charinfo.major = major(st.st_rdev);
                      charinfo.minor = minor(st.st_rdev);
                      charinfo.ino = st.st_ino;
                      charinfo.mode = st.st_mode & 07777;
                      charinfo.uid = st.st_uid;
                      charinfo.gid = st.st_gid;
                      charinfo.atime = time2ll(st.st_atime);
                      charinfo.mtime = time2ll(st.st_mtime);
                      charinfo.ctime = time2ll(st.st_ctime);
                      printcharinfo(list, &charinfo);

                      break;

      case S_IFBLK  : blockinfo.major = major(st.st_rdev);
                      blockinfo.minor = minor(st.st_rdev);
                      blockinfo.ino = st.st_ino;
                      blockinfo.mode = st.st_mode & 07777;
                      blockinfo.uid = st.st_uid;
                      blockinfo.gid = st.st_gid;
                      blockinfo.atime = time2ll(st.st_atime);
                      blockinfo.mtime = time2ll(st.st_mtime);
                      blockinfo.ctime = time2ll(st.st_ctime);
                      printblockinfo(list, &blockinfo);

                      break;

      default       : fprintf(stderr, "unknown type for entry %s\n", name);

                      return 1;
    }
  }

  closedir(dir);

  return 0;
}

int backup(CONFIG *config)
{
  struct stat st;
  int sockfd;
  LIBSSH2_SESSION *ssh;
  LIBSSH2_SFTP *sftp;
  char dest[8192];
  char sourcelistname[8192];
  LIBSSH2_SFTP_HANDLE *file;
  FILEINFOLIST *sourcelist;
  time_t t;
  unsigned long long llt;
  char listname[8192];
  int listfd;
  FILE *list;
  unsigned long dirnum;
  unsigned long filenum;
  int ret;
  char destlistname[8192];

  if (lstat(config->localdir, &st) < 0)
  {
    perror(config->localdir);

    return 1;
  }

  if (!serverconnect(config, &sockfd, &ssh, &sftp))
  {
    fprintf(stderr, "serverconnect failed\n");

    return 1;
  }

  if (!findlastlist(sftp, config, sourcelistname))
  {
    fprintf(stderr, "sftpdir not found\n");
    serverdisconnect(sockfd, ssh, sftp);

    return 1;
  }

  if (sourcelistname[0] == 0)
    sourcelist = NULL;
  else
  {
    file = libssh2_sftp_open(sftp, sourcelistname, LIBSSH2_FXF_READ, 0);

    if (file == NULL)
    {
      fprintf(stderr, "libssh2_sftp_open failed\n");
      serverdisconnect(sockfd, ssh, sftp);

      return 1;
    }

    if ((sourcelist = loadfileinfolist(file)) == NULL)
    {
      fprintf(stderr, "loadfileinfolist failed\n");
      libssh2_sftp_close(file);
      serverdisconnect(sockfd, ssh, sftp);

      return 1;
    }

    if (libssh2_sftp_close(file) != 0)
    {
      fprintf(stderr, "libssh2_sftp_close failed\n");
      serverdisconnect(sockfd, ssh, sftp);

      return 1;
    }
  } 

  time(&t);
  llt = time2ll(t);
 
  sprintf(dest, "%s/%llu", config->sftpdir, llt);

  if (libssh2_sftp_mkdir(sftp, dest, LIBSSH2_SFTP_S_IRUSR|LIBSSH2_SFTP_S_IWUSR|LIBSSH2_SFTP_S_IXUSR) != 0)
  {
    fprintf(stderr, "libssh2_sftp_mkdir failed\n");
    freefileinfolist(sourcelist);
    serverdisconnect(sockfd, ssh, sftp);

    return 1;
  }

  strcpy(listname, "/tmp/kissback.XXXXXX");

  if ((listfd = mkstemp(listname)) < 0)
  {
    perror("mkstemp");
    freefileinfolist(sourcelist);
    serverdisconnect(sockfd, ssh, sftp);

    return 1;
  }

  if ((list = fdopen(listfd, "w")) == NULL)
  {
    perror(listname);
    freefileinfolist(sourcelist);
    serverdisconnect(sockfd, ssh, sftp);

    return 1;
  }

  dirnum = 0;
  filenum = 0;
  ret = backupdir(config, config->localdir, config->localdir, st.st_dev, sourcelist, llt, sftp, dest, list, &dirnum, &filenum);
  fclose(list);
  freefileinfolist(sourcelist);

  if (ret != 0)
  {
    serverdisconnect(sockfd, ssh, sftp);

    return ret;
  }

  if (lstat(listname, &st) < 0)
  {
    perror(listname);
    serverdisconnect(sockfd, ssh, sftp);

    return 1;
  }

  if ((listfd = open(listname, O_RDONLY)) < 0)  
  {
    perror("open");
    serverdisconnect(sockfd, ssh, sftp);

    return 1;
  }

  sprintf(destlistname, "%s/%llu.lst", config->sftpdir, llt);
  file = libssh2_sftp_open(sftp, destlistname, LIBSSH2_FXF_WRITE|LIBSSH2_FXF_CREAT, LIBSSH2_SFTP_S_IRUSR|LIBSSH2_SFTP_S_IWUSR);

  if (file == NULL)
  {
    fprintf(stderr, "libssh2_sftp_open failed\n");
    close(listfd);
    serverdisconnect(sockfd, ssh, sftp);

    return 1;
  }

  if (!copytosftp(listfd, file, st.st_size))
  {
    fprintf(stderr, "unable to copy %s to %s\n", listname, destlistname);
    libssh2_sftp_close(file);
    close(listfd);
    serverdisconnect(sockfd, ssh, sftp);

    return 1;
  }

  if (libssh2_sftp_close(file))
  {
    close(listfd);
    serverdisconnect(sockfd, ssh, sftp);
    fprintf(stderr, "libssh2_sftp_close failed\n");

    return 1;
  }

  close(listfd);
  serverdisconnect(sockfd, ssh, sftp);

  return 0;
}

int listdates(CONFIG *config)
{
  int sockfd;
  LIBSSH2_SESSION *ssh;
  LIBSSH2_SFTP *sftp;
  LIBSSH2_SFTP_HANDLE *dir;
  char name[1024];
  LIBSSH2_SFTP_ATTRIBUTES attributes;

  if (!serverconnect(config, &sockfd, &ssh, &sftp))
  {
    fprintf(stderr, "serverconnect failed\n");

    return 1;
  }

  dir = libssh2_sftp_opendir(sftp, config->sftpdir);

  if (dir == NULL)
  {
    fprintf(stderr, "sftpdir not found\n");
        serverdisconnect(sockfd, ssh, sftp);

    return 0;
  }

  while ((libssh2_sftp_readdir(dir, name, 1000, &attributes)) > 0)
  {
    if (endswith(name , ".lst"))
    {
      name[14] = 0;
      printf("%s\n", name);
    }
  }

  if (libssh2_sftp_closedir(dir) != 0)
  {
    serverdisconnect(sockfd, ssh, sftp);

    return 1;
  }

  serverdisconnect(sockfd, ssh, sftp);

  return 0;
}

int check(CONFIG *config)
{
  char listname[8192];
  int sockfd;
  LIBSSH2_SESSION *ssh;
  LIBSSH2_SFTP *sftp;
  LIBSSH2_SFTP_ATTRIBUTES attributes;
  LIBSSH2_SFTP_HANDLE *file;
  char tmplistname[8192];
  int tmplistfd;
  FILE *list;
  char name[8192];
  char info[8192];
  FILEINFO fileinfo;
  char backupname[8192];
  char sha256[8192];

  strcpy(listname, config->sftpdir);
  strcat(listname, "/");
  strcat(listname, config->date);
  strcat(listname, ".lst");

  if (!serverconnect(config, &sockfd, &ssh, &sftp))
  {
    fprintf(stderr, "serverconnect failed\n");

    return 1;
  }

  if (libssh2_sftp_lstat(sftp, listname, &attributes) != 0)
  {
    fprintf(stderr, "sftpdir or date not found\n");
    serverdisconnect(sockfd, ssh, sftp);

    return 1;
  }

  file = libssh2_sftp_open(sftp, listname, LIBSSH2_FXF_READ, 0);

  if (file == NULL)
  {
    fprintf(stderr, "libssh2_sftp_open failed\n");
    serverdisconnect(sockfd, ssh, sftp);

    return 1;
  }

  strcpy(tmplistname, "/tmp/kissback.XXXXXX");

  if ((tmplistfd = mkstemp(tmplistname)) < 0)
  {
    perror("mkstemp");
    libssh2_sftp_close(file);
    serverdisconnect(sockfd, ssh, sftp);

    return 1;
  }

  if (!copyfromsftp(file, tmplistfd, attributes.filesize))
  {
    fprintf(stderr, "unable to copy %s to %s\n", listname, tmplistname);
    close(tmplistfd);
    libssh2_sftp_close(file);
    serverdisconnect(sockfd, ssh, sftp);

    return 1;
  }
 
  close(tmplistfd);

  if (libssh2_sftp_close(file) != 0)
  {
    fprintf(stderr, "libssh2_sftp_close failed\n");
    serverdisconnect(sockfd, ssh, sftp);

    return 1;
  }

  if ((list = fopen(tmplistname, "r")) == NULL)
  {
    perror(tmplistname);
    serverdisconnect(sockfd, ssh, sftp);

    return 1;
  }

  for (;;)  
  {
    fgets(name, 8182, list);

    if (feof(list))
      break;

    name[strlen(name) - 1] = 0;

    fgets(info, 8182, list);

    if (feof(list))
    {
      fprintf(stderr, "info line missing\n");
      fclose(list);
      serverdisconnect(sockfd, ssh, sftp);
      
      return 1;
    }

    info[strlen(info) - 1] = 0;

    if (strncmp(info, "FILE", 4) != 0)
      continue;

    if (!parsefileinfo(info, &fileinfo))
    {
      fprintf(stderr, "parsefileinfo failed\n");
      fclose(list);
      serverdisconnect(sockfd, ssh, sftp);

      return 1;
    }

    if (strlen(fileinfo.backupname) < 4)
    { 
      fprintf(stderr, "backupname too short\n");
      fclose(list);
      serverdisconnect(sockfd, ssh, sftp);
        
      return 1;
    }

    if ((strlen(config->sftpdir) + strlen(fileinfo.backupname) + 10) > 8192)
    {
      fprintf(stderr, "backupname too long\n");
      fclose(list);
      serverdisconnect(sockfd, ssh, sftp);
 
      return 1;
    }

    strcpy(status, name);
    strcpy(backupname, config->sftpdir);
    strcat(backupname, "/");
    strcat(backupname, fileinfo.backupname);
    
    if (libssh2_sftp_lstat(sftp, backupname, &attributes) != 0)
    {
      fprintf(stderr, "libssh2_sftp_lstat failed for %s\n", backupname);
      fclose(list);
      serverdisconnect(sockfd, ssh, sftp);

      return 1;
    }

    if ((strcmp(&fileinfo.backupname[strlen(fileinfo.backupname) - 3], ".xz") != 0) && (attributes.filesize != fileinfo.size))
    {
      fprintf(stderr, "%s: size mismatchx\n", backupname);
      fclose(list);
      serverdisconnect(sockfd, ssh, sftp);

      return 1;
    }

    if (!config->sha256check)
      continue;

    file = libssh2_sftp_open(sftp, backupname, LIBSSH2_FXF_READ, 0);

    if (file == NULL)
    {
      fprintf(stderr, "libssh2_sftp_open access failed\n");
      fclose(list);
      serverdisconnect(sockfd, ssh, sftp);

      return 1;
    }

    if (strcmp(&fileinfo.backupname[strlen(fileinfo.backupname) - 3], ".xz") == 0)
    {
      if (!calcsha256lzmafromsftp(file, sha256))
      {
        fprintf(stderr, "%s: calcsizesha256 failed\n", backupname);
        libssh2_sftp_close(file);
        fclose(list);
        serverdisconnect(sockfd, ssh, sftp);

        return 1;
      }
    }
    else
    {
      if (!calcsha256fromsftp(file, fileinfo.size, sha256))
      {
        fprintf(stderr, "%s: calcsizesha256 failed\n", backupname);
        libssh2_sftp_close(file);
        fclose(list);
        serverdisconnect(sockfd, ssh, sftp);

        return 1;
      }
    }

    if (libssh2_sftp_close(file) != 0)
    {
      fprintf(stderr, "libssh2_sftp_close failed\n");
      fclose(list);
      serverdisconnect(sockfd, ssh, sftp);

      return 1;
    }

    if (strcmp(sha256, fileinfo.sha256) != 0)
    {
      fprintf(stderr, "%s: sha256 mismatch\n", backupname);
      fclose(list);
      serverdisconnect(sockfd, ssh, sftp);

      return 1;
    }
  }

  fclose(list);
  serverdisconnect(sockfd, ssh, sftp);

  return 0;
}

int restore(CONFIG *config)
{
  char listname[8192];
  int sockfd;
  LIBSSH2_SESSION *ssh;
  LIBSSH2_SFTP *sftp;
  LIBSSH2_SFTP_ATTRIBUTES attributes;
  LIBSSH2_SFTP_HANDLE *file;
  char tmplistname[8192];
  int tmplistfd;
  FILE *list;
  char name[8192];
  char destname[8192];
  char info[8192];
  DIRINFO dirinfo;
  FILEINFO fileinfo;
  LINKINFO linkinfo;
  FIFOINFO fifoinfo;
  CHARINFO charinfo;
  BLOCKINFO blockinfo;
  char backupname[8192];
  struct stat st;
  struct timeval tv[2];
  int destfd;
  char sha256[8192];

  strcpy(listname, config->sftpdir);
  strcat(listname, "/");
  strcat(listname, config->date);
  strcat(listname, ".lst");

  if (!serverconnect(config, &sockfd, &ssh, &sftp))
  {
    fprintf(stderr, "serverconnect failed\n");

    return 1;
  }

  if (libssh2_sftp_lstat(sftp, listname, &attributes) != 0)
  {
    fprintf(stderr, "sftpdir or date not found\n");
    serverdisconnect(sockfd, ssh, sftp);

    return 1;
  }

  file = libssh2_sftp_open(sftp, listname, LIBSSH2_FXF_READ, 0);

  if (file == NULL)
  {
    fprintf(stderr, "libssh2_sftp_open failed\n");
    serverdisconnect(sockfd, ssh, sftp);

    return 1;
  }

  strcpy(tmplistname, "/tmp/kissback.XXXXXX");

  if ((tmplistfd = mkstemp(tmplistname)) < 0)
  {
    perror("mkstemp");
    libssh2_sftp_close(file);
    serverdisconnect(sockfd, ssh, sftp);

    return 1;
  }

  if (!copyfromsftp(file, tmplistfd, attributes.filesize))
  {
    fprintf(stderr, "unable to copy %s to %s\n", listname, tmplistname);
    close(tmplistfd);
    libssh2_sftp_close(file);
    serverdisconnect(sockfd, ssh, sftp);

    return 1;
  }

  close(tmplistfd);

  if (libssh2_sftp_close(file) != 0)
  {
    fprintf(stderr, "libssh2_sftp_close failed\n");
    serverdisconnect(sockfd, ssh, sftp);

    return 1;
  }

  if ((list = fopen(tmplistname, "r")) == NULL)
  {
    perror(tmplistname);
    serverdisconnect(sockfd, ssh, sftp);

    return 1;
  }

  for (;;)  
  {
    fgets(name, 8182, list);

    if (feof(list))
      break;

    name[strlen(name) - 1] = 0;

    if ((strlen(config->localdir) + strlen(name) + 10) > 8192)
    {
      fprintf(stderr, "name too long\n");
      fclose(list);
      serverdisconnect(sockfd, ssh, sftp);

      return 1;
    }

    strcpy(destname, config->localdir);
    strcat(destname, "/");
    strcat(destname, name);

    if (stat(destname, &st) >= 0)
    {
      fprintf(stderr, "%s : file exists\n", destname);
      fclose(list);
      serverdisconnect(sockfd, ssh, sftp);

      return 1;
    }

    fgets(info, 8182, list);

    if (feof(list))
    {
      fprintf(stderr, "info line missing\n");
      fclose(list);
      serverdisconnect(sockfd, ssh, sftp);
      
      return 1;
    }

    info[strlen(info) - 1] = 0;

    if (strncmp(info, "DIR", 3) == 0)
    {
      if (!parsedirinfo(info, &dirinfo))
      {
        fprintf(stderr, "parsedirinfo failed\n");
        fclose(list);
        serverdisconnect(sockfd, ssh, sftp);

        return 1;
      }

      if (mkdir(destname, S_IRWXU) < 0)
      {
        perror(destname);
        fclose(list);
        serverdisconnect(sockfd, ssh, sftp);

        return 1;
      }

      if (lchown(destname, dirinfo.uid, dirinfo.gid) < 0)
      {
        perror(destname);
        fclose(list);
        serverdisconnect(sockfd, ssh, sftp);

        return 1;
      }

      if (chmod(destname, dirinfo.mode) < 0)
      {
        perror(destname);
        fclose(list);
        serverdisconnect(sockfd, ssh, sftp);

        return 1;
      }

      tv[0].tv_sec = ll2time(dirinfo.atime);
      tv[0].tv_usec = 0;
      tv[1].tv_sec = ll2time(dirinfo.mtime);
      tv[1].tv_usec = 0;

      if (lutimes(destname, tv) < 0)
      {
        perror(destname);
        fclose(list);
        serverdisconnect(sockfd, ssh, sftp);

        return 1;
      }
    }
    else if (strncmp(info, "FILE", 4) == 0)
    {
      if (!parsefileinfo(info, &fileinfo))
      {
        fprintf(stderr, "parsefileinfo failed\n");
        fclose(list);
        serverdisconnect(sockfd, ssh, sftp);

        return 1;
      }

      if (strlen(fileinfo.backupname) < 4)
      { 
        fprintf(stderr, "backupname too short\n");
        fclose(list);
        serverdisconnect(sockfd, ssh, sftp);
        
        return 1;
      }

      if ((strlen(config->sftpdir) + strlen(fileinfo.backupname) + 10) > 8192)
      {
        fprintf(stderr, "backupname too long\n");
        fclose(list);
        serverdisconnect(sockfd, ssh, sftp);

        return 1;
      }

      strcpy(status, name);
      strcpy(backupname, config->sftpdir);
      strcat(backupname, "/");
      strcat(backupname, fileinfo.backupname);
      file = libssh2_sftp_open(sftp, backupname, LIBSSH2_FXF_READ, 0);

      if (file == NULL)
      {
        fprintf(stderr, "libssh2_sftp_open access failed\n");
        fclose(list);
        serverdisconnect(sockfd, ssh, sftp);

        return 1;
      }

      if ((destfd = open(destname, O_WRONLY|O_CREAT, S_IRUSR|S_IWUSR)) < 0)
      {
        perror(destname);
        libssh2_sftp_close(file);
        fclose(list);
        serverdisconnect(sockfd, ssh, sftp);

        return 1;
      }

      if (strcmp(&fileinfo.backupname[strlen(fileinfo.backupname) - 3], ".xz") == 0)
      {
        if (!copylzmafromsftpsha256(file, destfd, sha256))
        {
          fprintf(stderr, "unable to copy %s to %s\n", name, destname);
          libssh2_sftp_close(file);
          close(destfd);
          fclose(list);
          serverdisconnect(sockfd, ssh, sftp);

          return 1;
        }
      }
      else
      {
        if (!copyfromsftpsha256(file, destfd, fileinfo.size, sha256))
        {
          fprintf(stderr, "unable to copy %s to %s\n", name, destname);
          libssh2_sftp_close(file);
          close(destfd);
          fclose(list);
          serverdisconnect(sockfd, ssh, sftp);

          return 1;
        }
      }

      close(destfd);

      if (libssh2_sftp_close(file) != 0)
      {
        fprintf(stderr, "libssh2_sftp_close failed\n");
        fclose(list);
        serverdisconnect(sockfd, ssh, sftp);

        return 1;
      }

      if (strcmp(sha256, fileinfo.sha256) != 0)
      {
        fprintf(stderr, "%s: sha256 mismatch\n", backupname);
        fclose(list);
        serverdisconnect(sockfd, ssh, sftp);

        return 1;
      }

      if (lchown(destname, fileinfo.uid, fileinfo.gid) < 0)
      {
        perror(destname);
        fclose(list);
        serverdisconnect(sockfd, ssh, sftp);

        return 1;
      }

      if (chmod(destname, fileinfo.mode) < 0)
      {
        perror(destname);
        fclose(list);
        serverdisconnect(sockfd, ssh, sftp);

        return 1;
      }

      tv[0].tv_sec = ll2time(fileinfo.atime);
      tv[0].tv_usec = 0;
      tv[1].tv_sec = ll2time(fileinfo.mtime);
      tv[1].tv_usec = 0;

      if (lutimes(destname, tv) < 0)
      {
        perror(destname);
        fclose(list);
        serverdisconnect(sockfd, ssh, sftp);

        return 1;
      }
    }
    else if (strncmp(info, "LINK", 4) == 0)
    {
      if (!parselinkinfo(info, &linkinfo))
      {
        fprintf(stderr, "parselinkinfo failed\n");
        fclose(list);
        serverdisconnect(sockfd, ssh, sftp);

        return 1;
      }

      if (symlink(linkinfo.targetname, destname) < 0)
      {
        perror(destname);
        fclose(list);
        serverdisconnect(sockfd, ssh, sftp);

        return 1;
      }

      if (lchown(destname, linkinfo.uid, linkinfo.gid) < 0)
      {
        perror(destname);
        fclose(list);
        serverdisconnect(sockfd, ssh, sftp);

        return 1;
      }

      tv[0].tv_sec = ll2time(linkinfo.atime);
      tv[0].tv_usec = 0;
      tv[1].tv_sec = ll2time(linkinfo.mtime);
      tv[1].tv_usec = 0;

      if (lutimes(destname, tv) < 0)
      {
        perror(destname);
        fclose(list);
        serverdisconnect(sockfd, ssh, sftp);

        return 1;
      }
    }
    else if (strncmp(info, "FIFO", 4) == 0)
    {
      if (!parsefifoinfo(info, &fifoinfo))
      {
        fprintf(stderr, "parsefifoinfo failed\n");
        fclose(list);
        serverdisconnect(sockfd, ssh, sftp);

        return 1;
      }

      if (mkfifo(destname, S_IRUSR|S_IWUSR) < 0)
      {
        perror(destname);
        fclose(list);
        serverdisconnect(sockfd, ssh, sftp);

        return 1;
      }

      if (lchown(destname, fifoinfo.uid, fifoinfo.gid) < 0)
      {
        perror(destname);
        fclose(list);
        serverdisconnect(sockfd, ssh, sftp);

        return 1;
      }

      if (chmod(destname, fifoinfo.mode) < 0)
      {
        perror(destname);
        fclose(list);
        serverdisconnect(sockfd, ssh, sftp);

        return 1;
      }

      tv[0].tv_sec = ll2time(fifoinfo.atime);
      tv[0].tv_usec = 0;
      tv[1].tv_sec = ll2time(fifoinfo.mtime);
      tv[1].tv_usec = 0;

      if (lutimes(destname, tv) < 0)
      {
        perror(destname);
        fclose(list);
        serverdisconnect(sockfd, ssh, sftp);

        return 1;
      }
    }
    else if (strncmp(info, "CHAR", 4) == 0)
    {
      if (!parsecharinfo(info, &charinfo))
      {
        fprintf(stderr, "parsecharinfo failed\n");
        fclose(list);
        serverdisconnect(sockfd, ssh, sftp);

        return 1;
      }

      if (mknod(destname, S_IFCHR|S_IRUSR|S_IWUSR, makedev(charinfo.major, charinfo.minor)) < 0)
      {
        perror(destname);
        fclose(list);
        serverdisconnect(sockfd, ssh, sftp);

        return 1;
      }

      if (lchown(destname, charinfo.uid, charinfo.gid) < 0)
      {
        perror(destname);
        fclose(list);
        serverdisconnect(sockfd, ssh, sftp);

        return 1;
      }

      if (chmod(destname, charinfo.mode) < 0)
      {
        perror(destname);
        fclose(list);
        serverdisconnect(sockfd, ssh, sftp);

        return 1;
      }

      tv[0].tv_sec = ll2time(charinfo.atime);
      tv[0].tv_usec = 0;
      tv[1].tv_sec = ll2time(charinfo.mtime);
      tv[1].tv_usec = 0;

      if (lutimes(destname, tv) < 0)
      {
        perror(destname);
        fclose(list);
        serverdisconnect(sockfd, ssh, sftp);

        return 1;
      }
    }
    else if (strncmp(info, "BLOCK", 4) == 0)
    {
      if (!parseblockinfo(info, &blockinfo))
      {
        fprintf(stderr, "parseblockinfo failed\n");
        fclose(list);
        serverdisconnect(sockfd, ssh, sftp);

        return 1;
      }

      if (mknod(destname, S_IFBLK|S_IRUSR|S_IWUSR, makedev(blockinfo.major, blockinfo.minor)) < 0)
      {
        perror(destname);
        fclose(list);
        serverdisconnect(sockfd, ssh, sftp);

        return 1;
      }

      if (lchown(destname, blockinfo.uid, blockinfo.gid) < 0)
      {
        perror(destname);
        fclose(list);
        serverdisconnect(sockfd, ssh, sftp);

        return 1;
      }

      if (chmod(destname, blockinfo.mode) < 0)
      {
        perror(destname);
        fclose(list);
        serverdisconnect(sockfd, ssh, sftp);

        return 1;
      }

      tv[0].tv_sec = ll2time(blockinfo.atime);
      tv[0].tv_usec = 0;
      tv[1].tv_sec = ll2time(blockinfo.mtime);
      tv[1].tv_usec = 0;

      if (lutimes(destname, tv) < 0)
      {
        perror(destname);
        fclose(list);
        serverdisconnect(sockfd, ssh, sftp);

        return 1;
      }
    }
  }

  rewind(list);

  for (;;)  
  {
    fgets(name, 8182, list);

    if (feof(list))
      break;

    name[strlen(name) - 1] = 0;

    if ((strlen(config->localdir) + strlen(name) + 10) > 8192)
    {
      fprintf(stderr, "name too long\n");
      fclose(list);
      serverdisconnect(sockfd, ssh, sftp);

      return 1;
    }

    strcpy(destname, config->localdir);
    strcat(destname, "/");
    strcat(destname, name);

    fgets(info, 8182, list);

    if (feof(list))
    {
      fprintf(stderr, "info line missing\n");
      fclose(list);
      serverdisconnect(sockfd, ssh, sftp);

      return 1;
    }

    info[strlen(info) - 1] = 0;

    if (strncmp(info, "DIR", 3) == 0)
    {
      if (!parsedirinfo(info, &dirinfo))
      {
        fprintf(stderr, "parsedirinfo failed\n");
        fclose(list);
        serverdisconnect(sockfd, ssh, sftp);

        return 1;
      }

      tv[0].tv_sec = ll2time(dirinfo.atime);
      tv[0].tv_usec = 0;
      tv[1].tv_sec = ll2time(dirinfo.mtime);
      tv[1].tv_usec = 0;

      if (lutimes(destname, tv) < 0)
      {
        perror(destname);
        fclose(list);
        serverdisconnect(sockfd, ssh, sftp);

        return 1;
      }
    }
  }

  fclose(list);
  serverdisconnect(sockfd, ssh, sftp);

  return 0;
}

int compact(CONFIG *config)
{
  int sockfd;
  LIBSSH2_SESSION *ssh;
  LIBSSH2_SFTP *sftp;
  LIBSSH2_SFTP_ATTRIBUTES attributes;
  LIBSSH2_SFTP_HANDLE *file;
  char sourcelistname[8192];
  char tmpsourcelistname[8192];
  char tmpsourcelistfd;
  FILE *sourcelist;
  time_t t;
  unsigned long long llt;
  char destination[8192];
  char listname[8192];
  int listfd;
  FILE *list;
  char name[8192];
  char info[8192];
  FILEINFO fileinfo;
  char sourcename[8192];
  unsigned long dirnum;
  unsigned long filenum;
  char destname[8192];
  struct stat st;
  char destlistname[8192];

  if (!serverconnect(config, &sockfd, &ssh, &sftp))
  {
    fprintf(stderr, "serverconnect failed\n");

    return 1;
  }

  if (!findlastlist(sftp, config, sourcelistname))
  {
    fprintf(stderr, "sftpdir not found\n");
    serverdisconnect(sockfd, ssh, sftp);

    return 1;
  }

  if (sourcelistname[0] == 0)
  {
    fprintf(stderr, "no previous backup found\n");
    serverdisconnect(sockfd, ssh, sftp);

    return 1;
  }

  if (libssh2_sftp_lstat(sftp, sourcelistname, &attributes) != 0)
  {
    fprintf(stderr, "libssh2_sftp_lstat failed\n");
    serverdisconnect(sockfd, ssh, sftp);

    return 1;
  }

  file = libssh2_sftp_open(sftp, sourcelistname, LIBSSH2_FXF_READ, 0);

  if (file == NULL)
  {
    fprintf(stderr, "libssh2_sftp_open failed\n");
    serverdisconnect(sockfd, ssh, sftp);

    return 1;
  }

  strcpy(tmpsourcelistname, "/tmp/kissback.XXXXXX");

  if ((tmpsourcelistfd = mkstemp(tmpsourcelistname)) < 0)
  {
    perror("mkstemp");
    libssh2_sftp_close(file);
    serverdisconnect(sockfd, ssh, sftp);

    return 1;
  }

  if (!copyfromsftp(file, tmpsourcelistfd, attributes.filesize))
  {
    fprintf(stderr, "unable to copy %s to %s\n", sourcelistname, tmpsourcelistname);
    close(tmpsourcelistfd);
    libssh2_sftp_close(file);

    return 1;
  }

  close(tmpsourcelistfd);

  if (libssh2_sftp_close(file) != 0)
  {
    fprintf(stderr, "libssh2_sftp_close failed\n");
    serverdisconnect(sockfd, ssh, sftp);

    return 1;
  }

  if ((sourcelist = fopen(tmpsourcelistname, "r")) == NULL)
  {
    perror(sourcelistname);
    serverdisconnect(sockfd, ssh, sftp);

    return 1;
  }

  time(&t);
  llt = time2ll(t);

  sprintf(destination, "%s/%llu", config->sftpdir, llt);

  if (libssh2_sftp_mkdir(sftp, destination, LIBSSH2_SFTP_S_IRUSR|LIBSSH2_SFTP_S_IWUSR|LIBSSH2_SFTP_S_IXUSR) != 0)
  {
    fprintf(stderr, "libssh2_sftp_mkdir failed\n");
    fclose(sourcelist);
    serverdisconnect(sockfd, ssh, sftp);

    return 1;
  }


  strcpy(listname, "/tmp/kissback.XXXXXX");

  if ((listfd = mkstemp(listname)) < 0)
  {
    perror("mkstemp");
    fclose(sourcelist);
    serverdisconnect(sockfd, ssh, sftp);

    return 1;
  }

  if ((list = fdopen(listfd, "w")) == NULL)
  {
    perror(listname);
    fclose(sourcelist);
    serverdisconnect(sockfd, ssh, sftp);

    return 1;
  }

  dirnum = 0;
  filenum = 0;

  for (;;)  
  {
    fgets(name, 8182, sourcelist);

    if (feof(sourcelist))
      break;

    fgets(info, 8182, sourcelist);

    if (feof(sourcelist))
    {
      fprintf(stderr, "info line missing\n");
      fclose(list);
      fclose(sourcelist);
      serverdisconnect(sockfd, ssh, sftp);
      
      return 1;
    }

    fputs(name, list);

    if (strncmp(info, "FILE", 4) != 0)
    {
      fputs(info, list);

      continue;
    }

    name[strlen(name) - 1] = 0;
    info[strlen(info) - 1] = 0;

    if (!parsefileinfo(info, &fileinfo))
    {
      fprintf(stderr, "parsefileinfo failed\n");
      fclose(list);
      fclose(sourcelist);
      serverdisconnect(sockfd, ssh, sftp);

      return 1;
    }

    if ((strlen(config->sftpdir) + strlen(fileinfo.backupname) + 10) > 8192)
    {
      fprintf(stderr, "backupname too long\n");
      fclose(list);
      fclose(sourcelist);
      serverdisconnect(sockfd, ssh, sftp);
 
      return 1;
    }

    strcpy(sourcename, config->sftpdir);
    strcat(sourcename, "/");
    strcat(sourcename, fileinfo.backupname);

    if (filenum == 1000)
    {
      dirnum++;
      filenum = 0;
    }

    if (filenum == 0)
    {
      sprintf(destname, "%s/%06lu", destination, dirnum);

      if (libssh2_sftp_mkdir(sftp, destname, LIBSSH2_SFTP_S_IRUSR|LIBSSH2_SFTP_S_IWUSR|LIBSSH2_SFTP_S_IXUSR) != 0)
      {
        perror(destname);
        fclose(list);
        fclose(sourcelist);
        serverdisconnect(sockfd, ssh, sftp);

        return 1;
      }
    }

    sprintf(destname, "%s/%06lu/%03lu.dat", destination, dirnum, filenum);

    if (libssh2_sftp_rename(sftp, sourcename, destname))
    {
      perror(destname);
      fclose(list);
      fclose(sourcelist);
      serverdisconnect(sockfd, ssh, sftp);

      return 1;
    }

    sprintf(fileinfo.backupname, "%llu/%06lu/%03lu.dat", llt, dirnum, filenum);

    printfileinfo(list, &fileinfo);

    filenum++;
  }

  fclose(list);
  fclose(sourcelist);

  if (lstat(listname, &st) < 0)
  {
    perror(listname);
    serverdisconnect(sockfd, ssh, sftp);

    return 1;
  }

  if ((listfd = open(listname, O_RDONLY)) < 0)
  {
    perror("open");
    serverdisconnect(sockfd, ssh, sftp);

    return 1;
  }

  sprintf(destlistname, "%s/%llu.lst", config->sftpdir, llt);
  file = libssh2_sftp_open(sftp, destlistname, LIBSSH2_FXF_WRITE|LIBSSH2_FXF_CREAT, LIBSSH2_SFTP_S_IRUSR|LIBSSH2_SFTP_S_IWUSR);

  if (file == NULL)
  { 
    fprintf(stderr, "libssh2_sftp_open failed\n");
    close(listfd);
    serverdisconnect(sockfd, ssh, sftp);
    
    return 1;
  }

  if (!copytosftp(listfd, file, st.st_size))
  {
    fprintf(stderr, "unable to copy %s to %s\n", listname, destlistname);
    close(listfd);
    serverdisconnect(sockfd, ssh, sftp);

    return 1;
  }

  if (libssh2_sftp_close(file))
  {
    close(listfd);
    serverdisconnect(sockfd, ssh, sftp);
    fprintf(stderr, "libssh2_sftp_close failed\n");

    return 1;
  }

  close(listfd);
  serverdisconnect(sockfd, ssh, sftp);

  return 0;
}

void help()
{
  fprintf(stderr, "kissback version 0.2 - keep it simple stupid backup\n");
  fprintf(stderr, "Copyright (C) 2016 David Arendt\n");
  fprintf(stderr, "Website: http://prrescue.prnet.org/index.php/kissback\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "This program is free software: you can redistribute it and/or modify\n");
  fprintf(stderr, "it under the terms of the GNU General Public License as published by\n");
  fprintf(stderr, "the Free Software Foundation, either version 3 of the License, or\n");
  fprintf(stderr, "(at your option) any later version.\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "This program is distributed in the hope that it will be useful,\n");
  fprintf(stderr, "but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
  fprintf(stderr, "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n");
  fprintf(stderr, "GNU General Public License for more details.\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "You should have received a copy of the GNU General Public License\n");
  fprintf(stderr, "along with this program.  If not, see <http://www.gnu.org/licenses/>.\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "kissback -mode backup|listdates|check|restore|compact\n");
  fprintf(stderr, "         -sftpserver <server>\n");
  fprintf(stderr, "         -sftpport <port>\n");
  fprintf(stderr, "         -sftpfingerprint <fingerprint>\n");
  fprintf(stderr, "         -sftpuser <user>\n");
  fprintf(stderr, "         -sftpkey <private key file>\n");
  fprintf(stderr, "         -sftpdir <directory>\n");
  fprintf(stderr, "         -localdir <local directory>\n");
  fprintf(stderr, "         -compression <level 0-9 (recommended: 6)>\n");
  fprintf(stderr, "         -date <date>\n");
  fprintf(stderr, "         -sha256check\n");
}

int main(int argc, char *argv[])
{
  CONFIG config;
  int i;
  struct sigaction action;

  if (argc < 2)
  {
    help();

    return 1;
  }

  memset(&config, 0, sizeof(CONFIG));
  config.sftpport = 22;

  for (i = 1; i < argc; i++)
  {
    if (strcmp(argv[i], "-mode") == 0)
    {
      i++;

      if (i >= argc)
      {
        fprintf(stderr, "argument missing for -mode\n");
	
	return 1;
      }

      if (strcmp(argv[i], "backup") == 0)
        config.mode = MODE_BACKUP;
      else if (strcmp(argv[i], "listdates") == 0)
        config.mode = MODE_LISTDATES;
      else if (strcmp(argv[i], "check") == 0)
        config.mode = MODE_CHECK;
      else if (strcmp(argv[i], "restore") == 0)
        config.mode = MODE_RESTORE;
      else if (strcmp(argv[i], "compact") == 0)
        config.mode = MODE_COMPACT;
      else
      {
        fprintf(stderr, "unknown mode for -mode\n");

        return 1;
      }

    }
    else if (strcmp(argv[i], "-sftpserver") == 0)
    { 
      i++;
      
      if (i >= argc)
      { 
        fprintf(stderr, "argument missing for -sftpserver\n");
        
        return 1;
      }

      if (strlen(argv[i]) > 1000)
      { 
        fprintf(stderr, "argument too long for -sftpserver\n");

        return 1;
      }

      strcpy(config.sftpserver, argv[i]);
    }
    else if (strcmp(argv[i], "-sftpport") == 0)
    {
      i++;

      if (i >= argc)
      {
        fprintf(stderr, "argument missing for -sftpport\n");

        return 1;
      }

      config.sftpport = atoi(argv[i]);
    }
    else if (strcmp(argv[i], "-sftpfingerprint") == 0)
    {
      i++;

      if (i >= argc)
      {
        fprintf(stderr, "argument missing for -sftpfingerprint\n");

        return 1;
      }

      if (strlen(argv[i]) > 1000)
      {
        fprintf(stderr, "argument too long for -sftpfingerprint\n");

        return 1;
      }

      strcpy(config.sftpfingerprint, argv[i]);
    }
    else if (strcmp(argv[i], "-sftpuser") == 0)
    {
      i++;

      if (i >= argc)
      {
        fprintf(stderr, "argument missing for -sftpuser\n");

        return 1;
      }

      if (strlen(argv[i]) > 1000)
      {
        fprintf(stderr, "argument too long for -sftpuser\n");

        return 1;
      }

      strcpy(config.sftpuser, argv[i]);
    }
    else if (strcmp(argv[i], "-sftpkey") == 0)
    { 
      i++;
      
      if (i >= argc)
      { 
        fprintf(stderr, "argument missing for -sftpkey\n");
        
        return 1;
      }
      
      if (strlen(argv[i]) > 1000)
      { 
        fprintf(stderr, "argument too long for -sftpkey\n");
        
        return 1;
      }
      
      strcpy(config.sftpkey, argv[i]);
    }
    else if (strcmp(argv[i], "-sftpdir") == 0)
    {
      i++;
    
      if (i >= argc)
      {
        fprintf(stderr, "argument missing for -sftpdir\n");
    
        return 1;
      }

      if (strlen(argv[i]) > 1000)
      {
        fprintf(stderr, "argument too long for -sftpdir\n");

        return 1;
      }

      strcpy(config.sftpdir, argv[i]);
    }
    else if (strcmp(argv[i], "-localdir") == 0)
    {
      i++;
    
      if (i >= argc)
      {
        fprintf(stderr, "argument missing for -localdir\n");
    
        return 1;
      }

      if (strlen(argv[i]) > 1000)
      {
        fprintf(stderr, "argument too long for -localdir\n");

        return 1;
      }

      strcpy(config.localdir, argv[i]);
    }
    else if (strcmp(argv[i], "-compression") == 0)
    {
      i++;

      if (i >= argc)
      {
        fprintf(stderr, "argument missing for -compression\n");

        return 1;
      }

      config.compression = atoi(argv[i]);

      if (config.compression > 9)
      {
        fprintf(stderr, "invalid compression level for -compression\n");

        return 1;
      }
    }
    else if (strcmp(argv[i], "-date") == 0)
    {
      i++;
    
      if (i >= argc)
      {
        fprintf(stderr, "argument missing for -date\n");
    
        return 1;
      }

      if (strlen(argv[i]) > 1000)
      {
        fprintf(stderr, "argument too long for -date\n");

        return 1;
      }

      strcpy(config.date, argv[i]);
    }
    else if (strcmp(argv[i], "-sha256check") == 0)
      config.sha256check = 1;
    else
    { 
      fprintf(stderr, "unknown argument %s\n", argv[i]);
        
      return 1;
    }
  }

  if (config.sftpserver[0] == 0)
  {
    fprintf(stderr, "argument -sftpserver missing\n");

    return 1;
  }

  if (config.sftpuser[0] == 0)
  { 
    fprintf(stderr, "argument -sftpuser missing\n");
    
    return 1;
  }

  if (config.sftpkey[0] == 0)
  {
    fprintf(stderr, "argument -sftpkey missing\n");

    return 1;
  }

  if (config.sftpdir[0] == 0)
  { 
    fprintf(stderr, "argument -sftpdir missing\n");
    
    return 1;
  }

  if (config.mode == MODE_BACKUP)
  {
    if (config.localdir[0] == 0)
    {
      fprintf(stderr, "argument -localdir missing\n");

      return 1;
    }
  }
  else if (config.mode == MODE_CHECK)
  {
    if (config.date[0] == 0)
    {
      fprintf(stderr, "argument -date missing\n");

      return 1;
    }
  }
  else if (config.mode == MODE_RESTORE)
  {
    if (config.localdir[0] == 0)
    {
      fprintf(stderr, "argument -localdir missing\n");

      return 1;
    }

    if (config.date[0] == 0)
    {
      fprintf(stderr, "argument -date missing\n");

      return 1;
    }
  }

  strcpy(status, "processing file list...");
  memset(&action, 0, sizeof(action));
  action.sa_sigaction = &showstatus;
  action.sa_flags = SA_SIGINFO;
  sigaction(SIGUSR1, &action, NULL);

  if (config.mode == MODE_BACKUP)
    return backup(&config);
  else if (config.mode == MODE_LISTDATES)
    return listdates(&config);
  else if (config.mode == MODE_CHECK)
    return check(&config);
  else if (config.mode == MODE_RESTORE)
    return restore(&config);
  else if (config.mode == MODE_COMPACT)
    return compact(&config);

  return 1;
}
